/* Code to emulate the Daylight Drawing Library and turn them into the
appropriate Python callbacks.

  Implementation:

Most functions are commands.  They are simply converted into Python
callbacks for the current rendering object.

Some are queries.

COPYRIGHT:
  The interface definitions can be found from Daylight's v451 distribution
  in contrib/src/libdu/depict/dl_stubs.c which states:

  / **********************************************************************
    * This source code is public domain, and may be freely distributed,   *
    * modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO    *
    * WARRANTY OF ANY KIND.                                               *
    ********************************************************************** /

"Public domain" being what it is, this library is LGPL'ed.


*/

#define DEBUG_CODE
#ifdef DEBUG_CODE
/* #define DB(x) x */
#include <stdlib.h>
#define DB(x) if (getenv("DEPICT") && getenv("DEPICT")[0] == '1') {x;}
#include <stdio.h>
#else
#define DB(x)
#endif


#include "dl_defs.h"
#include "dt_smiles.h"
#include "dt_depict.h"
#include "Python.h"

PyObject *callback = NULL;

/* Used to store traceback information */
PyObject *ptype;
PyObject *pvalue;
PyObject *ptraceback;


#define PRINT_ERROR							\
    if (PyErr_Occurred()) {						\
      /* Daylight cannot handle these errors so give some feedback */	\
      PyErr_Print();							\
      PyErr_Clear();							\
    }

#define CHECK_METHOD(obj)				\
  if (!obj) {						\
    PRINT_ERROR						\
  } else {						\
    Py_DECREF(obj);					\
  }

#define CALL_VOID(fctn, method)				\
void fctn(void)						\
{							\
  PyObject *obj;					\
  DB(printf(method "\n"));				\
							\
  obj = PyObject_CallMethod(callback, method, NULL);	\
  CHECK_METHOD(obj)					\
}

#define CALL_INTEGER(fctn, method)			\
void fctn(int val)					\
{							\
  PyObject *obj;					\
  DB(printf(method "\n"));				\
							\
  obj = PyObject_CallMethod(callback, method, "i", val);\
  CHECK_METHOD(obj)					\
}

#define GET_INTEGER(fctn, method, deflt)		\
int fctn(void)						\
{							\
  PyObject *obj;					\
  int val = deflt;					\
  DB(printf(method "\n"));				\
							\
  obj = PyObject_CallMethod(callback, method, NULL);	\
  if (!obj) {						\
    PRINT_ERROR;					\
    return val;						\
  }							\
  if (!PyArg_GetInt(obj, &val)) {			\
    /* if it fails, it still uses the default */	\
    PRINT_ERROR;					\
  }							\
  Py_DECREF(obj);					\
  return val;						\
}


/** called before drawing begins */
CALL_VOID(dl_beginscreen, "beginscreen")

/** called when drawing complete */
CALL_VOID(dl_endscreen, "endscreen")

/** called when output buffer should be flushed */
CALL_VOID(dl_flush, "flush")


/** sets desired number of cells per drawing */
CALL_INTEGER(dl_setcellcount, "setcellcount")

/** sets rows/cols in grid (setcallcount alternative) */
/* Not needed??? Isn't referenced in our v451 depict library */
/* void dl_setcellrc(void) */

/** returns actual number of cells */
GET_INTEGER(dl_cellcount, "cellcount", 1)

/** returns current number of rows in cells */
GET_INTEGER(dl_cellrows, "cellrows", 1)

/** returns current number of rows in cell grid */
GET_INTEGER(dl_cellcols, "cellcols", 1)

/** sets current number of cell */
CALL_INTEGER(dl_setcell, "setcell")

/** returns current cell number */
GET_INTEGER(dl_cell, "cell", 0)

/** sets current ga (e.g. DL_GA_DEFAULT) */
/*  This is the "graphical attribute."  */
CALL_INTEGER(dl_setga, "setga")

GET_INTEGER(dl_ga, "ga", DL_GA_DEFAULT)

/** returns color mode (e.g. DL_COLORMODE_MONOCHROME) */
/* Appears these can be:   DL_COLORMODE_*
   MONOCHROME, GRAYSCALE, STDCOLOR, BOW, WOB, COW, COB */
GET_INTEGER(dl_colormode, "colormode", DL_COLORMODE_COB)

/** sets line width (e.g. DL_LINEWIDTH_THIN) */
/* valid line widths appear to be: DL_LINEWIDTH_*
   DEFAULT, THIN, NORMAL, WIDE, EXTRAWIDE */
CALL_INTEGER(dl_setlinewidth, "setlinewidth")

/** sets line style (e.g. DL_LINESTYLE_SOLID) */
/* valid line styles appear to be: DL_LINESTYLE_*
   DEFAULT, SOLID, DOTTED, DASHED */
CALL_INTEGER(dl_setlinestyle, "setlinestyle")

/** sets actual window size in pixels */
void dl_setpixels(DL_PIXEL *pix)
{
  PyObject *obj;
  DB(printf("setpixels h=%d v=%d", pix->h, pix->v));

  obj = PyObject_CallMethod(callback, "setpixels", "(ii)", pix->h, pix->v);
  CHECK_METHOD(obj);
}
/** returns actual window size */
DL_PIXEL dl_pixels(void)
{
  DL_PIXEL pix = {0, 0};
  PyObject *obj;
  DB(printf("pixels\n"));

  obj = PyObject_CallMethod(callback, "pixels", NULL);
  if (!obj) {
    PRINT_ERROR;
    return pix;
  }
  if (!PyArg_ParseTuple(obj, "ii", &pix.h, &pix.v)) {
    PRINT_ERROR;
  }
  Py_DECREF(obj);
  return pix;
}

/** sets drawing bounds to given limits */
void dl_setbounds(DL_RECT *rect)
{
  PyObject *obj;
  DB(printf("setbounds xmin=%f ymin=%f xmax=%f ymax=%f\n",
	    rect->xmin, rect->ymin, rect->xmax, rect->ymax));

  obj = PyObject_CallMethod(callback, "setbounds", "(ff)(ff)",
			    rect->xmin, rect->ymin, rect->xmax, rect->ymax);
  CHECK_METHOD(obj);
			    
}

/** sets drawing bounds to fit limits at 1:1 aspect */
void dl_fitbounds(DL_RECT *rect)
{
  PyObject *obj;
  DB(printf("fitbounds xmin=%f ymin=%f xmax=%f ymax=%f\n",
	    rect->xmin, rect->ymin, rect->xmax, rect->ymax));

  obj = PyObject_CallMethod(callback, "fitbounds", "(ff)(ff)",
			    rect->xmin, rect->ymin, rect->xmax, rect->ymax);
  CHECK_METHOD(obj);
}

/** sets/unsets whether setbounds/fixbounds is active */
CALL_INTEGER(dl_fixbounds, "fixbounds")

/** returns current drawing bounds */
DL_RECT dl_bounds(void)
{
  DL_RECT rect = {0.0, 0.0, 1.0, 1.0};
  PyObject *obj;
  DB(printf("bounds\n"));

  obj = PyObject_CallMethod(callback, "bounds", NULL);
  if (!obj) {
    PRINT_ERROR;
    return rect;
  }

  /* Otherwise there is the confusing message:
     SystemError: new style getargs format but argument is not a tuple */
  if (!PyTuple_Check(obj)) {
    PyErr_SetString(PyExc_TypeError, 
		    "bounds must return (xmin, ymin),(xmax, ymin)");
    PRINT_ERROR;
    return rect;
  }

  if (!PyArg_ParseTuple(obj, "(ff)(ff)", &rect.xmin, &rect.ymin,
			&rect.xmax, &rect.ymax )) {
    PRINT_ERROR;
  }
  Py_DECREF(obj);
  return rect;
}

/** I assume this sets the aspect ratio ??? */
void dl_setaspect(float aspect)
{
  PyObject *obj;
  DB(printf("setaspect %f\n", aspect));

  obj = PyObject_CallMethod(callback, "fitbounds", "f", aspect);
  CHECK_METHOD(obj);
}
/** Get the current aspect ratio ??? */
float dl_aspect(void)
{
  PyObject *obj;
  float val = 1.0;
  DB(printf("aspect\n"));

  obj = PyObject_CallMethod(callback, "aspect", NULL);
  if (!PyArg_Parse(obj, "f", &val)) {
    PRINT_ERROR;
  } else {
    Py_DECREF(obj);
  }
  return val;
}

/** returns physical aspect ratio */
/* Not in the v451 header files!! */
float dl_paspect(void)
{
  PyObject *obj;
  float val = 1.0;
  DB(printf("paspect\n"));

  obj = PyObject_CallMethod(callback, "paspect", NULL);
  if (!PyArg_Parse(obj, "f", &val)) {
    PRINT_ERROR;
  } else {
    Py_DECREF(obj);
  }
  return val;
}

/** (re)sets current drawing point */
void dl_moveto(DL_POINT *point)
{
  PyObject *obj;
  DB(printf("moveto x=%f y=%f\n", point->x, point->y));

  obj = PyObject_CallMethod(callback, "moveto", "ff", point->x, point->y);
  CHECK_METHOD(obj);
}

/** draws from current point to the given point */
void dl_drawto(DL_POINT *point)
{
  PyObject *obj;
  DB(printf("drawto x=%f y=%f\n", point->x, point->y));

  obj = PyObject_CallMethod(callback, "drawto", "ff", point->x, point->y);
  CHECK_METHOD(obj);
}

/** draws a rectangle */
void dl_drawbox(DL_RECT *rect)
{
  PyObject *obj;
  DB(printf("drawbox xmin=%f ymin=%f xmax=%f ymax=%f\n",
	    rect->xmin, rect->ymin, rect->xmax, rect->ymax));
  obj = PyObject_CallMethod(callback, "drawbox", "(ff)(ff)",
			    rect->xmin, rect->ymin, rect->xmax, rect->ymax);
  CHECK_METHOD(obj);
}

/** draws an oval */
void dl_drawoval(DL_RECT *rect)
{
  PyObject *obj;
  DB(printf("drawoval xmin=%f ymin=%f xmax=%f ymax=%f\n",
	 rect->xmin, rect->ymin, rect->xmax, rect->ymax));
  obj = PyObject_CallMethod(callback, "drawoval", "(ff)(ff)",
			    rect->xmin, rect->ymin, rect->xmax, rect->ymax);
  CHECK_METHOD(obj);
}

/* internal debugging */
void print_label(DL_LABEL *label)
{
  DB(
  printf("text=%s, sizes=%s, super=%s, tag=%c(%d)%c(%d)%c(%d)%c(%d) "
	 "loc.x=%f loc.y=%f size=%f angle=%f\n",
	 label->text, label->sizes, label->super,
	 label->tag[0], label->tag[0], label->tag[1], label->tag[1],
	 label->tag[2], label->tag[2], label->tag[3], label->tag[3],
	 label->loc.x, label->loc.y,
	 label->size, label->angle);
  )
}

/* should do this as a class.  Don't know how.  Instead, use dict */
static PyObject* label_to_dict(DL_LABEL *label)
{
  PyObject *ob = PyDict_New();
  if (!ob) {
    return NULL;
  }
  PyMapping_SetItemString(ob, "text", PyString_FromString(label->text));
  PyMapping_SetItemString(ob, "sizes", PyString_FromString(label->sizes));
  PyMapping_SetItemString(ob, "super", PyString_FromString(label->super));
  PyMapping_SetItemString(ob, "tag", PyString_FromStringAndSize(label->tag,4));
  PyMapping_SetItemString(ob, "loc", Py_BuildValue("(ff)", label->loc.x,
						   label->loc.y));
  PyMapping_SetItemString(ob, "size", PyFloat_FromDouble(label->size));
  PyMapping_SetItemString(ob, "angle", PyFloat_FromDouble(label->angle));
  return ob;
}


/** returns bounds of text label of given size */
DL_RECT dl_labelbounds(DL_LABEL *label)
{
  PyObject *py_label, *obj;
  DL_RECT rect = {0.0, 0.0, 1.0, 1.0};
  DB(printf("labelbounds "));
  DB(print_label(label));

  py_label = label_to_dict(label);
  if (!py_label) {
    PRINT_ERROR;
    return rect;
  }
  obj = PyObject_CallMethod(callback, "labelbounds", "O", py_label);
  Py_DECREF(py_label);

  if (!obj) {
    PRINT_ERROR;
    return rect;
  }

  /* Otherwise there is the confusing message:
     SystemError: new style getargs format but argument is not a tuple */
  if (!PyTuple_Check(obj)) {
    PyErr_SetString(PyExc_TypeError, 
		    "bounds must return (xmin, ymin),(xmax, ymin)");
    PRINT_ERROR;
    return rect;
  }
  if (!PyArg_ParseTuple(obj, "(ff)(ff)", &rect.xmin, &rect.ymin,
			&rect.xmax, &rect.ymax)) {
    PRINT_ERROR;
  }
  Py_DECREF(obj);

  return rect;
}

/** draws text label of given size */
void dl_drawlabel(DL_LABEL *label)
{
  PyObject *py_label, *obj;

  DB(printf("drawlabel "));
  DB(print_label(label));
  py_label = label_to_dict(label);
  if (!py_label) {
    PRINT_ERROR;
    return;
  }
  obj = PyObject_CallMethod(callback, "drawlabel", "O", py_label);
  Py_DECREF(py_label);
  CHECK_METHOD(obj)
  
}


/** void draws text label in rectangle of given size */
void dl_fitlabel(DL_LABEL *label, DL_RECT *box)
{
  PyObject *py_label, *obj;

  DB(printf("fitlabel\n  label => "));
  DB(print_label(label));
  DB(printf("   box => xmin=%f ymin=%f, xmax=%f ymax=%f\n",
	    box->xmin, box->ymin, box->xmax, box->ymax));

  py_label = label_to_dict(label);
  if (!py_label) {
    PRINT_ERROR;
    return;
  }
  obj = PyObject_CallMethod(callback, "fitlabel", "O((ff)(ff))",
			    py_label, box->xmin, box->ymin,
			    box->xmax, box->ymax);
  Py_DECREF(py_label);
  CHECK_METHOD(obj);
}

/*********************** STDCOLOR implementation **************/

/* Colors are stored in a Python list */
PyObject *colors = NULL; /* later init'ed to Py_None */

/** returns the number of standard colors */
int dl_stdcolor_count(void)
{
  long len;
  /* check for colors init??? */
  DB(printf("color_count\n"));
  len = PyObject_Length(colors);
  if (len == -1) {
    return NULL;
  }
  return len;
}

/* initializes the standard color table */
int dl_stdcolor_init(void *dpy, int screen)
{
  DB(printf("color_init dpy=%p screen=%d", dpy, screen));
  if (colors == NULL) {
    return 0;
  }
  return 1;
}

/* returns the standard color name of an entry */
char *dl_stdcolor_name(int index)
{
  PyObject *obj, *name;
  char *s;

  DB(printf("color_name %d\n", index));

  obj = PySequence_GetItem(colors, index);
  if (obj == NULL) {
    return "";  /* what should be done in this case ??? */
  }

  /* should have a "name" field */
  name = PyObject_GetAttrString(obj, "name");
  if (name == NULL) {
    s = "";
  } else {
    /* get the string */
    s = PyString_AsString(name);
  }
  DB(printf("Got color_name:%s:\n", s));
  return s;
}

/* returns the defined RGB value of an entry */
int dl_stdcolor_def(int index, int *red, int *green, int *blue)
{
  PyObject *obj, *rgb;
  int r, g, b;

  DB(printf("color_def %d\n", index));
  obj = PySequence_GetItem(colors, index);

  if (obj == NULL) {
    goto error0;
  }

  /* should have a "rgb" field with 3 items */
  rgb = PyObject_GetAttrString(obj, "rgb");
  if (rgb == NULL) {
    goto error0;
  }

  if (!PyArg_ParseTuple(rgb, "iii", &r, &g, &b)) {
    goto error0;
  }
  if (r<0) r=0; else { if (r>255) r=255; }
  if (g<0) g=0; else { if (g>255) g=255; }
  if (b<0) b=0; else { if (b>255) b=255; }

  *red = r;
  *green = g;
  *blue = b;

  DB(printf("color_def: %d == %d %d %d\n", index, r, g, b));
  return 1;


  error0:
  PyErr_Clear();
  *red = *blue = *green = 0.0;
  return 0;
}

/* returns the actual RGB value used */
int dl_stdcolor_act(int index, int *red, int *green, int *blue)
{
  return dl_stdcolor_def(index, red, green, blue);
}

/* returns a standard color index given a name */
int dl_stdcolor_index(char *name)
{
  PyObject *pyname, *obj, *objname;
  long i, len;
  int result, retval = -1;

  DB(printf("color_index %s.", name));

  /* turn into a string for comparisons */
  pyname = PyString_FromString(name);

  len = PyObject_Length(colors);
  if (len == -1) {
    return -1;
  }
  for (i=0; i<len; i++) {
    obj = PySequence_GetItem(colors, i);
    objname = PyObject_GetAttrString(obj, "name");
    if (PyObject_Cmp(pyname, objname, &result) != -1 &&
	result == 0) {
      retval = i;
      break;
    }
  }
  Py_DECREF(pyname);
  DB(printf("stdcolor_index:%s: %d\n", name, retval));
  return retval;
}

/* returns a "pixel" of color for display (X11 only) */
/* Huh ??? */
int dl_stdcolor_pixel(int index)
{
  DB(printf("color_pixel %d", index));
  return index;
}


/* Takes the depiction, the callback and the color list */
/*static*/ PyObject *
pydepict(PyObject *self, PyObject *args)
{
  int dep;
  if (!PyArg_ParseTuple(args, "iOO", &dep, &callback, &colors)) {
    return NULL;
  }

  return PyInt_FromLong(dt_depict(dep));
}

/* wrapper for dt_getcoord since I don't know how to do it 
   correctly with SWIG */
/*static*/ PyObject *
getcoord(PyObject *self, PyObject *args)
{
  int dep, atom;
  float x, y, z;
  if (!PyArg_ParseTuple(args, "ii", &dep, &atom)) {
    return NULL;
  }
  if (dt_getcoord(dep, atom, &x, &y, &z)) {
    return Py_BuildValue("fff", x, y, z);
  }
  /* I wonder what went wrong */
  PyErr_SetString(PyExc_IndexError, "atom not found in depiction");
  return NULL;
}

/** Now merged into dayswig_python.c **/
/* static PyMethodDef daydepict_methods[] = { */
/*   {"depict", pydepict, METH_VARARGS}, */
/*   {"getcoord", getcoord, METH_VARARGS}, */
/* }; */

/* void */
/* initdaydepict_python() */
/* { */
/*   (void) Py_InitModule("daydepict_python", daydepict_methods); */
/* } */

