package require Dayswig_tcl
package provide libdaytcl 1.0

########################################################################
# FILE: du_atomic_mass.c
#
# SYNOPSIS:
#
#   float du_atomic_mass(dt_Handle atom);
#
#	Returns the atomic mass of the given atom, if known.  The mass is
#	determined as follows:
#
#	  - If the atom object's atomic number has been explicitely set
#	    (e.g. [2H], [12C], [15N] in a SMILES, or via dt_setweight()),
#	    and the exact isotopic mass is known for that weight, the exact
#	    isotopic mass is returned.
#
#	  - If the atom object's atomic number is not explicitely set (e.g.
#	    "C", "[CH4]", or dt_setweight(atom, 0)), then the average
#	    weight of naturally-occuring isotopes is returned.
#
########################################################################

proc du_atomic_mass {atom} {
    upvar \#0 du_atomic_mass..aaw aaw;
    upvar \#0 du_atomic_mass..iaw iaw;

    # initialize aaw -- average atomic weight data

    set aaw {
	0.000 1.008 4.0026 6.94 9.01218 10.81 12.011
	14.0067 15.9994 18.99846 20.17 22.98977 24.305 26.9815
	28.086 30.9738 32.06 35.453 39.948 39.1 40.08 44.9559
	47.9 50.941 51.996 54.938 55.847 58.9332 58.71 63.543
	65.38 69.72 72.59 74.9216 78.96 79.904 83.8 85.467
	87.62 88.9059 91.22 92.9064 95.94 98.9062 101.07
	102.9055 106.4 107.868 112.4 114.82 118.69 121.75 127.6
	126.9045 131.3 132.9055 137.34 138.9055 140.12 140.9077
	144.24 145. 150.4 151.96 157.25 158.9254 162.5 164.9303
	167.26 168.9342 173.04 174.97 178.49 180.947 183.85
	186.2 190.2 192.22 195.09 196.9665 200.59 204.37 207.2
	208.9806 209. 210. 222. 223. 226. 227. 232.0381
	231.0359 238.029 237.0408 244. 243. 247. 247. 251.
	254.09 257.08 258.098 259.10 260.1 261.1 262.11
    };

    # initialize iaw - isotopic atomic weight data

    array set iaw \
	[list \
	     [DX_ATN_H]:1 1.007825 \
	     [DX_ATN_H]:2 2.0140 \
	     [DX_ATN_H]:3 3.01605 \
	     [DX_ATN_C]:12 12.000000 \
	     [DX_ATN_C]:13 13.003355 \
	     [DX_ATN_C]:14 14.003241 \
	     [DX_ATN_C]:15 15.010599 \
	     [DX_ATN_C]:16 16.014701 \
	     [DX_ATN_N]:14 14.003074 \
	     [DX_ATN_N]:15 15.000108 \
	     [DX_ATN_O]:16 15.994915 \
	     [DX_ATN_O]:17 16.999131 \
	     [DX_ATN_O]:18 17.999160 \
	     [DX_ATN_F]:19 18.998403 \
	     [DX_ATN_Si]:28 27.976927 \
	     [DX_ATN_Si]:29 28.976495 \
	     [DX_ATN_Si]:30 29.973770 \
	     [DX_ATN_P]:31 30.973762 \
	     [DX_ATN_S]:32 31.972070 \
	     [DX_ATN_S]:33 32.971456 \
	     [DX_ATN_S]:34 33.967866 \
	     [DX_ATN_Cl]:35 34.968852 \
	     [DX_ATN_Cl]:37 36.965903 \
	     [DX_ATN_Br]:79 78.918336 \
	     [DX_ATN_Br]:81 80.916289 \
	     [DX_ATN_I]:127 126.904473 \
	    ];

    # initialize the real procedure
    #
    #   du_atomic_mass() -- return atomic mass for given atom
    #

    proc du_atomic_mass {atom} {
	upvar \#0 du_atomic_mass..aaw aaw;
	upvar \#0 du_atomic_mass..iaw iaw;

	set atno [dt_number $atom];

	if {$atno < 0 || $atno >= [llength $aaw]} {
	    return 0;
	}

	#  Return average mass if no isotope specified.
	
	set atwt [dt_weight $atom];
	if {0 == $atwt} {
	    return [lindex $aaw $atno];
	}
	
	# Return isotopic mass for common/high-defect isotopes.
	
	if {[info exists iaw($atno:$atwt)]} {
	    return $iaw($atno:$atwt);
	}
	
	# Otherwise, float integer atomic mass and return it
	
	return $atwt;
    }

    # return the real procedure's value
    return [du_atomic_mass $atom];
}
