package require Dayswig_tcl
package provide libdaytcl 1.0

########################################################################
#  du_elimination.c
#
#  Given two terminal atoms in different molecules, create a new molecule
#  by eliminating the given atoms.  This copies isotopic information across,
#  but not isomeric information (should probably fix this for sake of example).
#
#  This had the small problem that it added bonds to the old molecules
#  rather than to the new ones.
#
#  Revision: 7 March 1995 (JY) ... Now handles stereochemistry.
########################################################################

#
#  du_elimination(a0,a1) -- join two molecules by eliminating atoms
#


proc du_elimination {a0 a1} {
    set m0    [dt_parent $a0];
    set m1    [dt_parent $a1];
    set newa0 [NULL_OB];
    set newa1 [NULL_OB];

    # Insist on two atoms from different molecules.

    if {[TYP_ATOM] != [dt_type $a0]
	|| [TYP_ATOM] != [dt_type $a1]
	|| $m0 == $m1} {
	return [NULL_OB];
    }

    # Create new molecule.

    set mol [dt_alloc_mol];
    dt_mod_on $mol;

    # Copy atoms from input molecules, identifying new atoms with the old.

    foreach m [list $m0 $m1] {
	foreach atom [du_stream_to_list [dt_stream $m [TYP_ATOM]]] {
	    set new($atom) [dt_addatom $mol [dt_number $atom] [dt_imp_hcount $atom]];
	    dt_setweight $new($atom) [dt_weight $atom];
	    dt_setcharge $new($atom) [dt_charge $atom];
	}
    }

    # Check that given atoms were found. This isn't really needed.

    if { ! [info exists new($a0)] || ! [info exists new($a1)]} {
	dt_dealloc $mol;
	error "Rats!";
    }

    # Copy bonds from input molecules.

    foreach m [list $m0 $m1] {
	foreach bond [du_stream_to_list [dt_stream $m [TYP_BOND]]] {
	    pset end0 end1 [du_stream_to_list [dt_stream $bond [TYP_ATOM]]];
	    set new($bond) [dt_addbond $new($end0) $new($end1) [dt_bondtype $bond]];
	}
    }

    #
    # For each atom, determine chiral value
    # with dt_chival, or skip if not possible.
    #
    foreach atom [concat \
		      [du_stream_to_list [dt_stream $m0 [TYP_ATOM]]] \
		      [du_stream_to_list [dt_stream $m1 [TYP_ATOM]]]] {
	#
	# determine the number of attached atoms, we are only concerned
	# with tetrahedral chirality, so more than one implicit hydrogen
	# or not exactly 4 bonds and we are done.
	#
	set imp_hcount [dt_imp_hcount $atom];
	if {$imp_hcount > 1} {
	    continue;
	}
	set bonds [du_stream_to_list [dt_stream $atom [TYP_BOND]]];
	if {4 != $imp_hcount + [llength $bonds]} {
	    continue;
	}

	#
	# Build a bond sequence with which to determine chirality.
	# If exactly one implicit hydrogen, we must insert a placeholder H
	# with dt_isohydro() at start of bond-sequences.
	#
	if {1 == $imp_hcount} {
	    set bondseq [du_list_to_sequence [concat [dt_isohydro] $bonds]];
	} else {
	    set bondseq [du_list_to_sequence $bonds];
	}
	set chival [dt_chival $atom $bondseq];
	dt_dealloc $bondseq;

	#
	# if the chirality is undefined, then we are done
	#
	if {-1 == $chival || [DX_CHI_NONE] == $chival} {
	    continue;
	}

	#
	# copy the chirality from the parent
	#
	set newbonds {};
	foreach bond $bonds {
	    lappend newbonds $new($bond);
	}
	if {1 == $imp_hcount} {
	    set bondseq [du_list_to_sequence [concat [dt_isohydro] $newbonds]];
	} else {
	    set bondseq [du_list_to_sequence $newbonds];
	}
	dt_setchival $new($atom) $bondseq $chival;
	dt_dealloc $bondseq;
    }

    # Look for cis/trans in m0 and m1; copy to mol.

    foreach bond [concat \
		      [du_stream_to_list [dt_stream $m0 [TYP_BOND]]] \
		      [du_stream_to_list [dt_stream $m1 [TYP_BOND]]]] {
	if {2 == [dt_bondtype $bond]} {
	    pset atom0 atom1 rest [du_stream_to_list [dt_stream $bond [TYP_ATOM]]];
	    set bond0 [du_other_bond $atom0 $bond];
	    set bond1 [du_other_bond $atom1 $bond];
	    if {"$bond0" != {} && "$bond1" != {}} {
		set dboval [dt_dbo $bond $bond0 $bond1];
		if {[DX_CHI_NO_DBO]!= $dboval} {
		    dt_setdbo $new($bond) $new($bond0) $new($bond1) $dboval;
		}
	    }
	}
    }
   
    # Join molecules and eliminate given atoms.

    set end0 [du_connected_atom $new($a0)];
    set end1 [du_connected_atom $new($a1)];
    dt_dealloc $new($a0);
    dt_dealloc $new($a1);
    dt_addbond $end0 $end1 [DX_BTY_SINGLE];

    # Analyze molecule.

    if { ! [dt_mod_off $mol]} {
	return [NULL_OB];
    }

    return $mol;
}

