package require Dayswig_tcl
package provide libdaytcl 1.0

#
#  du_part.c
#
#
#
#  du_findparts(mol) -- find disconnected parts of `mol'; return part count
#
#  0 is returned on error.
#


proc du_findparts {mol} {
    upvar \#0 du_findparts..apart ap;
    upvar \#0 du_findparts..bpart bp;

    # Delete any old part arrays.

    catch {unset ap};
    catch {unset bp};
    
    # Return 0 on invalid request.

    if {[TYP_MOLECULE] != [dt_type $mol]} {
	return 0;
    }

    # Fill part arrays and return number of parts.

    return [du_findparts..fill_parts $mol ap bp];
}

#
#  du_part(ob) -- return part number of object `ob'
#
#  If 'ob' is an atom or bond, returns part number; otherwise, returns -1.
#
#  du_findparts() must be called before this routine.
#
#  The part number returned here is 0-origin (-1 returned on error).
#  If 1-origin numbering desired, don't subtract 1 before returning.
#

proc du_part {ob} {
    upvar \#0 du_findparts..apart ap;
    upvar \#0 du_findparts..bpart bp;

    # If parts not computed, return -1.

    if { ! [info exists ap]} {
	return -1;
    }

    # Return appropriate part number.

    if {[dt_type $ob] == [TYP_ATOM]} {
	return $ap([dt_uid $ob]);
    } elseif {[dt_type $ob] == [TYP_BOND]} {
	return $bp([dt_uid $ob]);
    }

    # Return -1 on invalid request.

    return -1;
}

#
#  fill_parts(mol) -- fill part arrays and return part count
#

proc du_findparts..fill_parts {mol aname bname} {
    upvar $aname ap;
    upvar $bname bp;

    set np 1;

    # Loop over atoms, setting parts with a depth-first-search.
    
    foreach atom [du_atoms $mol] {
	if { ! [info exists ap([dt_uid $atom])]} {
	    du_findparts..set_part_dfs $atom ap bp $np;
	    incr np;
	}
    }

    return [expr $np - 1];
}

#
#  set_part_dfs(atom) -- set part of atom, and connected atoms, to npart
#

proc du_findparts..set_part_dfs {atom aname bname n} {
    upvar $aname ap;
    upvar $bname bp;
    
    set as [list $atom]
    
    while {[llength $as] > 0} {
	set a [lindex $as 0];
	set as [lrange $as 1 end];
	set ap([dt_uid $a]) $n;

	foreach b [du_bonds $a] {
	    if {[info exists bp([dt_uid $b])]} {
		continue;
	    }
	    set bp([dt_uid $b]) $n;
	    set to [dt_xatom $a $b];
	    if {[info exists ap([dt_uid $to])]} {
		continue;
	    }
	    lappend as $to;
	}
    }
}

