package require Dayswig_tcl
package provide libdaytcl 1.0

#
# split a molecule into its disconnected parts, preserving
# all information present.
#
# returns a tcl list of molecules.
#

proc du_split {mol} {

    for {set i 1} {$i <= [du_findparts $mol]} {incr i} {
	set m($i) [dt_alloc_mol];
	dt_mod_on $m($i);
	lappend mols $m($i);
    }

    # copy each atom into its appropriate molecule

    foreach atom [du_atoms $mol] {
	set new($atom) [dt_addatom $m([du_part $atom]) [dt_number $atom] [dt_imp_hcount $atom]];
	dt_setweight $new($atom) [dt_weight $atom];
	dt_setcharge $new($atom) [dt_charge $atom];
    }

    # copy each bond into its appropriate molecule

    foreach bond [du_bonds $mol] {
	pset end0 end1 [du_atoms $bond];
	set new($bond) [dt_addbond $new($end0) $new($end1) [dt_bondtype $bond]];
    }

    # copy chirality specifications

    foreach atom [du_atoms $mol] {
	#
	# determine the number of attached atoms, we are only concerned
	# with tetrahedral chirality, so more than one implicit hydrogen
	# or not exactly 4 bonds and we are done.
	#
	set imp_hcount [dt_imp_hcount $atom];
	if {$imp_hcount > 1} {
	    continue;
	}
	set bonds [du_bonds $atom];
	if {4 != $imp_hcount + [llength $bonds]} {
	    continue;
	}

	#
	# Build a bond sequence with which to determine chirality.
	# If exactly one implicit hydrogen, we must insert a placeholder H
	# with dt_isohydro() at start of bond-sequences.
	#
	if {1 == $imp_hcount} {
	    set bondseq [du_list_to_sequence [concat [dt_isohydro] $bonds]];
	} else {
	    set bondseq [du_list_to_sequence $bonds];
	}
	set chival [dt_chival $atom $bondseq];
	dt_dealloc $bondseq;

	#
	# if the chirality is undefined, then we are done
	#
	global DX_CHI_NONE
	if {-1 == $chival || $DX_CHI_NONE == $chival} {
	    continue;
	}

	#
	# copy the chirality from the parent
	#
	set newbonds {};
	foreach bond $bonds {
	    lappend newbonds $new($bond);
	}
	if {1 == $imp_hcount} {
	    set bondseq [du_list_to_sequence [concat [dt_isohydro] $newbonds]];
	} else {
	    set bondseq [du_list_to_sequence $newbonds];
	}
	dt_setchival $new($atom) $bondseq $chival;
	dt_dealloc $bondseq;
    }

    # copy cis/trans specifications

    global DX_CHI_NO_DBO
    foreach bond [du_bonds $mol] {
	if {2 == [dt_bondtype $bond]} {
	    pset atom0 atom1 rest [du_atoms $bond];
	    set bond0 [du_other_bond $atom0 $bond];
	    set bond1 [du_other_bond $atom1 $bond];
	    if {"$bond0" != {} && "$bond1" != {}} {
		set dboval [dt_dbo $bond $bond0 $bond1];
		if {$DX_CHI_NO_DBO != $dboval} {
		    dt_setdbo $new($bond) $new($bond0) $new($bond1) $dboval;
		}
	    }
	}
    }

    # be finished
    foreach x $mols {
	dt_mod_off $x;
    }

    # return the new molecule(s)

    return $mols;
}
